# -*- coding: utf-8 -*-
#
# Testing for pl
#
# ------------------------------------------------


# imports
# -------
import pytest
from flask_occam import validate, optional
from flask_occam import log
from flask_occam import transactional
from wtforms import Form, StringField, BooleanField, PasswordField, validators

from .fixtures import app, Item, logs, logger

from flask_occam.errors import ValidationError


# log
# ---
@log('test with {var} and {obj.__class__.__name__}')
def log_default(var, obj):
    pass


@log.info('test with {var} and {obj.__class__.__name__}')
def log_info(var, obj):
    pass


@log.error('test with {var} and {obj.__class__.__name__}')
def log_error(var, obj):
    pass


@app.route('/log/<int:var>')
@log.info('test with {var}')
def get_log(var):
    pass


class TestLogDecorator(object):

    def test_url_logging(self, client):
        client.get('/log/0')
        assert 'test with 0' in logs.getvalue()
        return

    def test_formatting(self):
        log_default(1, object())
        assert 'test with 1 and object' in logs.getvalue()

        log_info(2, object())
        assert 'test with 2 and object' in logs.getvalue()

        log_error(3, object())
        assert 'test with 3 and object' in logs.getvalue()
        return


# paginate
# --------
class TestPaginateDecorator(object):

    def test_integration(self, client, items):

        # paginated request
        response = client.get('/items')
        assert response.status_code == 206
        assert len(response.json) == 2
        assert 'limit=2&offset=2' in response.headers['Link']
        assert 'rel="next"' in response.headers['Link']
        assert 'rel="last"' in response.headers['Link']

        # no pagination needed
        response = client.get('/items?limit=2000&offset=2')
        assert 'Link' not in response.headers
        assert len(response.json) > 0
        assert response.status_code == 200
        return


# transactional
# -------------
class TestTransactionalDecorator(object):

    @transactional
    def create_item(self):
        item = Item.create(name='transaction-okay')
        return item.id

    @transactional
    def create_item_error(self):
        item = Item.create(name='transaction-fail')
        raise AssertionError
        return item.id

    def test_transactions(self):
        # normal
        ident = self.create_item()
        assert Item.get(ident) is not None

        # rollback
        try:
            self.create_item_error()
        except AssertionError:
            pass
        assert Item.get(name='transaction-fail') is None
        return


# validate tests
# --------------
class TestValidateDecorator(object):

    def test_integration(self, client):
        item = Item.create(name='validation-test').commit()

        # invalid
        response = client.put('/items/{}'.format(item.id), json=dict(
            name=1
        ))
        assert response.status_code == 422
        assert 'name' in response.json['errors']

        # valid
        response = client.put('/items/{}'.format(item.id), json=dict(
            name='test'
        ))
        assert response.status_code == 200
        return

    def test_validate_types(self):
        from flask import _request_ctx_stack
        _request_ctx_stack.pop()

        # function
        @validate(
            one=str, two=float,
            three=list,
            four=optional(dict(
                foo=str,
                bar=str,
            )),
            five=optional([str])
        )
        def test(one, two, three, four=None, five=None):
            pass

        # no exception
        test(one='test', two=1.5, three=[1])
        test(one='test', two=1.5, three=[1], four=dict(foo='foo', bar='bar'), five=['one', 'two'])

        # non-optional
        try:
            test(one=1, two='test', three=1)
            self.fail('ValueError not thrown')
        except ValueError as exc:
            message = str(exc)
            assert 'one:' in message
            assert 'two:' in message
            assert 'three:' in message

        # optional
        try:
            test(one='test', two=1.5, three=[1], four=dict(foo=1, bar='bar'), five=[1, 2])
            self.fail('ValueError not thrown')
        except ValueError as exc:
            message = str(exc)
            assert 'four:' in message
            assert 'five:' in message
        return

    def test_validate_validators(self):
        from flask import _request_ctx_stack
        _request_ctx_stack.pop()

        # function
        @validate(
            email=validators.Email(),
            check=optional(StringField('check', [
                validators.Length(min=3, max=10),
                validators.EqualTo('confirm')
            ])),
            confirm=optional(validators.Length(min=3, max=10))
        )
        def test(email, check, confirm):
            pass

        # raises error
        with pytest.raises(ValueError):
            test(email='test', check='test', confirm='test')

        # raises error
        with pytest.raises(ValueError):
            test(email='a@b.com', check='t', confirm='t')

        # raises error
        with pytest.raises(ValueError):
            test(email='a@b.com', check='test', confirm='aaaa')

        # no error
        assert test(email='a@b.com', check='test', confirm='test') is None
        return

    def test_validate_form(self):
        from flask import _request_ctx_stack
        _request_ctx_stack.pop()

        # form
        class ValidateForm(Form):
            string = StringField('String', [
                validators.DataRequired(),
                validators.Length(min=3, max=10)
            ])
            boolean = BooleanField('Boolean', [
                validators.DataRequired(),
            ])


        @validate(ValidateForm)
        def test(string, boolean):
            pass

        # raises error
        with pytest.raises(ValueError):
            test(string='a', boolean='test')

        # no error
        assert test(string='test', boolean=True) is None
        return

    def test_validate_mixed(self):
        from flask import _request_ctx_stack
        _request_ctx_stack.pop()

        # function
        @validate(
            email=validators.Email(),
            one=dict(length=validators.Length(min=3)),
            two=float
        )
        def test(email, one, two):
            pass

        # no error
        test(email='a@b.com', one=dict(length='test'), two=1.5)

        # nested error
        with pytest.raises(ValueError):
            test(email='a@b.com', one=dict(length='te'), two=1.5)
        return





# -*- coding: utf-8 -*-

import os
import sys

TESTS = os.path.dirname(os.path.realpath(__file__))
BASE = os.path.realpath(os.path.join(TESTS, '..'))
RESOURCES = os.path.join(TESTS, 'resources')
SANDBOX = os.path.join(TESTS, 'sandbox')




# -*- coding: utf-8 -*-
#
# Testing for sqlalchemy extensions
#
# ------------------------------------------------


# imports
# -------
from .fixtures import db, Item


# tests
# -----
class TestModelMixin(object):

    def test_utils(self, client, items):
        # json
        data = items[0].json()
        assert 'id' in data
        assert 'name' in data
        assert 'archived' in data

        # count
        count = Item.count()
        assert count > 0
        return

    def test_crud(self, client, items):
        # create
        item = Item.create(name='mixin-crud', archived=False).commit()

        # upsert
        items = Item.upsert([
            dict(name='mixin-crud', archived=True),
            dict(name='mixin-crud-new', archived=False)
        ])
        db.session.commit()
        updated = Item.find(name='mixin-crud')
        assert len(updated) == 1
        assert updated[0].archived is True
        created = Item.get(name='mixin-crud-new')
        assert created is not None

        # get
        ident = item.id
        item = Item.get(ident)
        assert item.name == 'mixin-crud'

        # get with alternate filter
        item = Item.get(name='mixin-crud')
        assert item.id == ident
        return

        # find
        items = Item.find(archived=False)
        assert len(items) > 3
        assert all([not x.archived for x in items])

        # update
        item = Item.get(name='mixin-crud')
        item.update(name='mixin-crud-update', archived=True).commit()
        updated = Item.get(name='mixin-crud-update')
        assert updated.id == item.id

        # delete
        item.delete()
        item = Item.get(item.id)
        assert item is None
        return



# -*- coding: utf-8 -*-

__pkg__ = 'Flask-Occam'
__url__ = 'https://github.com/bprinty/Flask-Occam'
__info__ = 'Flask extension for simplifying REST API development.'
__author__ = 'Blake Printy'
__email__ = 'bprinty@gmail.com'
__version__ = '0.1.0'


from .plugin import Occam, Blueprint         ## noqa

from .decorators import optional             ## noqa
from .decorators import validate             ## noqa
from .decorators import transactional        ## noqa
from .decorators import paginate             ## noqa
from .decorators import log                  ## noqa

from .handlers import ActionHandler          ## noqa
from .handlers import QueryHandler           ## noqa

from .errors import ValidationError          ## noqa



# -*- coding: utf-8 -*-
#
# Common utilities for app
#
# ------------------------------------------------


# imports
# -------
import sys
import yaml
import inspect
import logging
from functools import wraps
from wtforms.fields.core import UnboundField, Field
from wtforms import validators, Form, StringField
from werkzeug.exceptions import ExpectationFailed
from werkzeug.datastructures import MultiDict, ImmutableMultiDict
from flask import request, current_app, Response, has_request_context

from .errors import ValidationError

if sys.version_info[0] >= 3:
    unicode = str

try:
    from flask_login import current_user
except ImportError:
    current_user = None


# logging
# -------
class log(object):
    """
    Decorator for adding default logging functionality
    to api methods.
    """

    def __init__(self, msg, level=None):
        self.msg = msg
        if current_app:
            default = current_app.config['OCCAM_LOG_DEFAULT_LEVEL']
        else:
            default = 'info'
        self.level = default if not level else level
        return

    def __call__(self, func):
        # config logger to use
        if current_app:
            logger = getattr(current_app.logger, self.level)
        else:
            logger = getattr(logging, self.level)

        # inner function to format message
        @wraps(func)
        def inner(*args, **kwargs):
            data = kwargs.copy()
            varnames = list(inspect.signature(func).parameters.keys())
            data.update(dict(zip(varnames, args)))
            data = {k: v for k, v in data.items() if k not in ['cls', 'self']}
            data.setdefault(current_app.config['OCCAM_LOG_USER_FORMAT'], current_user)
            data.setdefault('kwargs', kwargs)
            logger(self.msg.format(**data))
            return func(*args, **kwargs)
        return inner

    @classmethod
    def debug(cls, msg):
        """
        Log provided message with level DEBUG.
        """
        return cls(msg, level='debug')

    @classmethod
    def info(cls, msg):
        """
        Log provided message with level INFO.
        """
        return cls(msg, level='info')

    @classmethod
    def warning(cls, msg):
        """
        Log provided message with level WARNING.
        """
        return cls(msg, level='warning')

    @classmethod
    def error(cls, msg):
        """
        Log provided message with level ERROR.
        """
        return cls(msg, level='error')

    @classmethod
    def critical(cls, msg):
        """
        Log provided message with level CRITICAL.
        """
        return cls(msg, level='critical')


# pagination
# ----------
def paginate(**options):
    """
    Automatically force request pagination for endpoints
    that shouldn't return all items in the database directly.
    If this decorator is used, ``limit`` and ``offset`` request
    arguments are automatically included in the request. The
    burden is then on developers to do something with those
    ``limit`` and ``offset`` arguments. An example request header
    set by this decorator is as follows:

    .. code-block:: text

        Link: <https://localhost/items?limit=50&offset=50>; rel="next",
              <https://localhost/items?limit=50&offset=500>; rel="last"

    Args:
        limit (int): Number of entries to limit a query by.
        total (int, callable): Number or callable for determining
            the total number of records that can be returned
            for the request. This is used in determining the
            pagination header.
    """
    if 'total' not in options:
        raise AssertionError(
            '`@paginate` decorator requires `total=` parameter '
            'for determining total number of records to paginate. '
            'See the documentation for more details.')

    def decorator(func):

        @wraps(func)
        def inner(*args, **kwargs):

            # only paginate on get requests
            if request.method != 'GET':
                return func(*args, **kwargs)

            # format parameters
            limit = request.args.get('limit', options.get('limit'))
            offset = int(request.args.get('offset', options.get('offset', 0)))
            total = options['total']() if callable(options['total']) else options['total']
            url = options.get('url', request.base_url)

            # config request parameters
            request.args = request.args.copy()
            request.args.setdefault('limit', limit)
            request.args.setdefault('offset', offset)

            # if no need to paginate, return without setting headers
            if limit is None:
                return func(*args, **kwargs)
            limit = int(limit)

            # add next page link
            headers = {}
            next_page = '<{}?limit={}&offset={}>'.format(url, limit, offset + limit)
            headers['Link'] = '{}; rel="next"'.format(next_page)

            # add last page link and header
            if options['total'] is not None:
                total = options['total']() if callable(options['total']) else options['total']
                last_page = '<{}?limit={}&offset={}>'.format(url, limit, offset + limit)
                headers['Link'] += ', {}; rel="last"'.format(last_page)
                headers['X-Total-Count'] = str(total)

            # call the function and create response
            response = func(*args, **kwargs)

            # if a specific response has already been crafted, use it
            if isinstance(response, Response):
                return response

            # normalize response data
            if not isinstance(response, tuple):
                response = [response]
            response = list(response)
            if hasattr(response[0], 'json'):
                content_length = len(response[0].json)
            else:
                content_length = len(response[0])
            if len(response) == 1:
                response.append(200)
            if len(response) == 2:
                response.append({})

            # if the response data is equal to the pagination, it's
            # truncated and needs updated headers/status
            if content_length == limit:
                response[1] = 206
                response[2].update(headers)

            return tuple(response)
        return inner
    return decorator


# db-related
# ----------
def transactional(func):
    """
    Decorator for wrapping transactional support around
    a request handler or API method. By wrapping a method
    in ``@transactional``, any exception thrown will force
    a session rollback. Otherwise, the session will be
    committed.
    """
    @wraps(func)
    def inner(*args, **kwargs):
        if 'sqlalchemy' not in current_app.extensions:
            raise AssertionError('Cannot use @transactional if Flask-SQLAlchemy extension has not been registered!')
        db = current_app.extensions['sqlalchemy'].db
        try:
            ret = func(*args, **kwargs)
            db.session.commit()
            return ret
        except Exception as e:
            db.session.rollback()
            raise e
    return inner


# validation
# ----------
class OptionalType:
    """
    Proxy for making a type optional. Used
    internally within the module.
    """

    def __init__(self, typ):
        self.type = typ
        return

    def __call__(self, item):
        return self.type(item)


def optional(field):
    """
    Replace DataRequired validtion scheme with Optional
    validation scheme.

    Args:
        field (type, Field): Field to process validation options for.
    """
    # parse types as inputs
    if isinstance(field, (type, list, tuple, dict)):
        return OptionalType(field)

    # handle validators directly
    elif not isinstance(field, (Field, UnboundField)):
        return StringField(field.__class__.__name__, [validators.Optional(), field])

    # handle wtform input class
    else:
        if len(field.args) < 2:
            raise AssertionError('Validator must have associated name and validation scheme (2 arguments)')
        name = field.args[0]
        checks = list(filter(
            lambda x: not isinstance(x, (validators.DataRequired, validators.InputRequired)),
            field.args[1]
        ))
        checks.insert(0, validators.Optional())
        obj = field.field_class(name, checks)
        return obj


def create_validator(contract):
    """
    Factory pattern for creating custom validator in the style
    of WTForms. This function takes the contract needed to validate
    against and returns a ``Form`` object that can perform
    the validation according to validation rules in the contract.

    Args:
        contract (dict): Dictionary with key value mappings for
                         payload contract to validate.
    """

    class ContractValidator(object):
        """
        Custom validator class that can handle both
        wtforms validators and nested type validation.
        """
        type_contract = {}
        form_contract = {}
        for key in contract:
            field = contract[key]
            if isinstance(field, (type, list, tuple, dict, OptionalType)):
                type_contract[key] = field
            elif isinstance(field, (Field, UnboundField)):
                form_contract[key] = field
            else:
                form_contract[key] = StringField(key, [
                    validators.DataRequired(),
                    field
                ])

        def __init__(self, data):
            self.data = data
            self.errors = {}
            return

        def check_type(self, actual, expected):
            # missing parameter (checked in api methods)
            if actual is None:
                return True

            # types
            if isinstance(expected, OptionalType):
                expected = expected.type

            if isinstance(expected, type):
                if expected == str:
                    return isinstance(actual, (str, unicode))
                else:
                    return isinstance(actual, expected)

            # dictionary of types
            elif isinstance(expected, dict):
                if not isinstance(actual, dict):
                    return False
                Validator = create_validator(expected)
                form = Validator(MultiDict(actual))
                return form.validate()

            # list of objects (expected structure should
            # have one entry representing shape of elements)
            elif isinstance(expected, list):
                if not isinstance(actual, list):
                    return False
                if isinstance(expected[0], type):
                    for entry in actual:
                        if not self.check_type(entry, expected[0]):
                            return False
                else:
                    Validator = create_validator(expected[0])
                    for entry in actual:
                        form = Validator(MultiDict(entry))
                        if not form.validate():
                            return False

            return True

        def check_form(self, actual, expected):
            class CustomValidator(Form):
                pass
            for key in expected:
                setattr(CustomValidator, key, expected[key])
            form = CustomValidator(MultiDict(actual))
            valid = form.validate()
            self.errors.update(form.errors)
            return valid

        def validate(self):
            """
            Run validation for composite form object.
            """
            valid = True

            # process types in contract
            if self.type_contract:
                for key in self.type_contract:
                    expect = self.type_contract[key]

                    # check for required fields
                    if key not in self.data and not isinstance(expect, OptionalType):
                        self.errors[key] = ['This field is required.']
                        valid = False

                    # check type and nested data
                    else:
                        tcheck = self.check_type(self.data.get(key), expect)
                        valid &= tcheck
                        if not tcheck:
                            typ = expect if not isinstance(expect, OptionalType) else expect.type
                            self.errors[key] = ['Invalid type. Expecting `{}`.'.format(typ)]

            # process form validators in contract
            if self.form_contract:
                valid &= self.check_form(self.data, self.form_contract)

            return valid

    return ContractValidator


def validate(*vargs, **vkwargs):
    """
    Validate payload data inputted to request handler or
    API method. This decorator can take any arbitrary
    data structure and determine rules to validate inputs with.
    It can be run within the context of a flask request, or
    for an api method directly. See the examples below for
    details.

    .. note:: In a request context, a ``ValidationError`` will be
              thrown (HTTP 422 Unprocessable Entity). Outside of a
              request, a ``ValueError`` will be raised.

    Examples:

        Using types with API method:

        .. code-block:: python

            @validate(
                email=str,
                password=str
            )
            def login(email, password):
                return

        Using nested types (mixing types and validators):

        .. code-block:: python

            from wtforms import validators

            @blueprint.route('/login', methods=['POST'])
            @validate(
                data=dict(
                    email=validators.Email(),
                    password=str
                )
            )
            def login():
                return

        With pre-defined validation object:

        .. code-block:: python

            from wtforms import validators, Form, BooleanField
            from wtforms import StringField, PasswordField
            class LoginValidator(Form):
                email = StringField('Email Address', [
                    validators.DataRequired(),
                    validators.Email(),
                ])
                password = PasswordField('Password', [
                    validators.DataRequired(),
                    validators.Length(min=4, max=25)
                ])

            @blueprint.route('/login', methods=['POST'])
            @validate(LoginForm)
            def login():
                return

        Logic specified into decorator:

        .. code-block:: python

            @blueprint.route('/login', methods=['POST'])
            @validate(
                email = StringField('Email Address', [
                    validators.DataRequired(),
                    validators.Email(),
                ])
                password = PasswordField('Password', [
                    validators.DataRequired(),
                    validators.Length(min=4, max=25)
                ])
            )
            def login():
                return
    """
    # validator class
    if len(vargs):
        Validator = vargs[0]

    # validator as keyword arguments
    elif len(vkwargs):
        Validator = create_validator(vkwargs)

    # incorrect arguments
    else:
        raise AssertionError('No validation rule supplied to @validate.')

    def decorator(func):
        @wraps(func)
        def inner(*args, **kwargs):
            in_request = has_request_context()

            # parse payload data from request or function arguments
            if in_request:
                data = request.json if request.json is not None else request.args
                if isinstance(data, ImmutableMultiDict):
                    data = data.to_dict()
            else:
                data = kwargs.copy()
                varnames = list(inspect.signature(func).parameters.keys())
                data.update(dict(zip(varnames, args)))
                data = {k: v for k, v in data.items() if k not in ['cls', 'self']}

            # check for empty payload
            if not data:
                if in_request:
                    raise ExpectationFailed('No request payload specified.')
                else:
                    raise AssertionError('No arguments specified.')

            # instantiate validator with data (changes based
            # on normal vs dynamic validator)
            if isinstance(Validator, Form):
                form = Validator(MultiDict(data))
            else:
                form = Validator(data=data)

            # run payload validation
            if not form.validate():
                if in_request:
                    raise ValidationError('Invalid request payload.', form.errors)
                else:
                    errors = yaml.dump(form.errors, indent=2)
                    errors = errors.replace('\n', '\n  ').replace('- ', '  - ') + '\n'
                    raise ValueError('Invalid arguments specified.\n\nErrors:\n\n  {}'.format(errors))

            # subset inputs by validators
            return func(*args, **kwargs)
        return inner
    return decorator



# -*- coding: utf-8 -*-
#
# Custom exceptions used throughout application
#
# ------------------------------------------------


# imports
# -------
from werkzeug.exceptions import HTTPException
from flask import jsonify


# errors
# ------
class ValidationError(HTTPException):
    """
    Custom error handler to manage request
    validation reporting.

    Args:
        message (str): String with error message.
        errors (dict): Dictionary with validation errors.
    """
    code = 422
    description = 'Unprocessable Entity'

    def __init__(self, message=None, errors=None):
        self.message = message if message is not None else self.description
        self.errors = errors
        return

    @staticmethod
    def handler(error):
        return jsonify({
            'msg': error.message,
            'errors': error.errors
        }), error.code



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Package setup
#
# ------------------------------------------------


# config
# ------
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages


config = __import__('flask_occam')
with open('requirements.txt', 'r') as reqs:
    requirements = map(lambda x: x.rstrip(), reqs.readlines())


test_requirements = [
    'pytest',
    'pytest-cov',
    'pytest-runner'
]


# files
# -----
with open('README.rst') as readme_file:
    readme = readme_file.read()


# exec
# ----
setup(
    name=config.__pkg__,
    version=config.__version__,
    description=config.__info__,
    long_description=readme,
    author=config.__author__,
    author_email=config.__email__,
    url=config.__url__,
    packages=find_packages(exclude=['tests']),
    license="MIT",
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    install_requires=requirements,
    keywords=[config.__pkg__, 'flask', 'rest', 'web', 'validation'],
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    tests_require=test_requirements
)



# -*- coding: utf-8 -*-
#
# Custom handlers for simplifying code related
# to url processing.
#
# ------------------------------------------------


# imports
# -------
from werkzeug.exceptions import NotFound


# handlers
# --------
class QueryHandler(object):
    """
    Handler for urls like ``/model/:id/:query``,
    where <query> represents a GET request. This
    allows developers to define individual methods
    for performing model-specific queries instead
    of needing to manage that complexity in the
    get/put/post methods.
    """

    def get(self, **kwargs):
        """
        Overwrite get endpoint for handler.

        Args:
            ident (int): Identifier for model.
            query (str): String for query to perform.
        """
        query = kwargs.pop('query', None)
        if not query or not hasattr(self, query):
            raise NotFound
        return getattr(self, query)(**kwargs)


class ActionHandler(object):
    """
    Handler for urls like ``/model/:id/:action``,
    where <action> represents a POST request. This
    allows developers to define individual methods
    for performing model-specific actions instead
    of needing to manage that complexity in the
    get/put/post methods.
    """

    def post(self, **kwargs):
        """
        Overwrite post endpoint for handler.

        Args:
            ident (int): Identifier for model.
            action (str): String for action to perform.
        """
        action = kwargs.pop('action', None)
        if not action or not hasattr(self, action):
            raise NotFound
        return getattr(self, action)(**kwargs)



# -*- coding: utf-8 -*-
#
# Plugin Setup
#
# ------------------------------------------------


# imports
# -------
import re
from functools import wraps
from flask import Flask, Blueprint, Response, jsonify
import types

from .converters import ModelConverter
from .mixins import ModelMixin
from .errors import ValidationError


# config
# ------
METHODS = [
    'GET',
    'HEAD',
    'POST',
    'PUT',
    'DELETE',
    'CONNECT',
    'OPTIONS',
    'TRACE',
    'PATCH'
]
DOCS = {}


# helpers
# -------
def document(obj):
    """
    Parse function or method and save documentation
    information in global DOCS data structure for
    auto-serving documentation.

    Args:
        obj (callable): Callable object with docstring.
    """
    global DOCS, METHODS
    if obj.__doc__:
        m = re.search(r"({})\s+(/.+?)\s".format('|'.join(METHODS)), obj.__doc__)
        if m:
            method, endpoint = m.group(1), m.group(2)
            if endpoint not in DOCS:
                DOCS[endpoint] = {}
            DOCS[endpoint][method] = obj.__doc__
    return


def autojsonify(func):
    """
    Automatically jsonify return payload if a dictionary
    is returned from the url handler. This removes some
    boilerplate associated with always needing to wrap
    json with `jsonify`.
    """
    @wraps(func)
    def _(*args, **kwargs):
        ret = func(*args, **kwargs)
        if isinstance(ret, tuple):
            if not isinstance(ret[0], Response):
                ret = list(ret)
                ret[0] = jsonify(ret[0])
                ret = tuple(ret)
        elif not isinstance(ret, Response):
            ret = jsonify(ret)
        return ret
    return _


def route(self, rule, **options):
    """
    Decorator for registering view function for a given URL rule.
    This ovverrides the default Flask route decorator to be able
    to handle class-based definitions, where endpoints are specified as
    class methods. With this, you can define routes with a function
    like you normally do in Flask:

    .. code-block:: python

        @app.route('/')
        def index():
            return 'Hello World'


    Or, you can define them with a class:

    .. code-block:: python

        @app.route('/test')
        class Test:
            def get(self):
                return 'GET /test', 200
            def post(self):
                return 'POST /test', 201


    This class-based definition of handlers helps with code organization
    and documentation efforts.
    """
    def decorator(obj):

        # function-based definition
        endpoint = options.pop("endpoint", obj.__name__)
        app = self.app if hasattr(self, 'app') else self
        if isinstance(obj, types.FunctionType):
            document(obj)
            app.add_url_rule(rule, endpoint, autojsonify(obj), **options)

        # class-based definition
        else:
            instance = obj()

            # document callable objects
            for name in dir(instance):
                func = getattr(instance, name)
                if name[0] != '_' and callable(func):
                    document(func)

            # parse/add url rules
            methods = options.pop("methods", METHODS)
            for meth in methods:
                handler = meth.lower()
                if hasattr(instance, handler):
                    endpoint = handler + '_' + obj.__name__.lower()
                    options['methods'] = [meth]
                    func = getattr(instance, handler)
                    app.add_url_rule(
                        rule, endpoint, autojsonify(func),
                        **options
                    )

        return obj
    return decorator


Blueprint.route = route


# plugin
# ------
class Occam(object):
    """
    Flask extension class for module, which sets up all flask-related
    capabilities provided by the module. This object can be initialized
    directly:

    .. code-block:: python

        from flask import Flask
        from flask_occam import Occam

        app = Flask(__name__)
        occam = Occam(app)

    Or lazily via factory pattern:

    .. code-block:: python

        occam = Occam()
        app = Flask(__name__)
        occam.init_app(app)

    For additional functionality, it is recommended that the extension
    be linked to the Flask-SQLAlchemy extension tied to the application:

    .. code-block:: python

        from flask_sqlalchemy import SQLAlchemy

        db = SQLAlchemy()
        occam = Occam(db)
        app = Flask(__name__)
        db.init_app(app)
        occam.init_app(app)
    """

    def __init__(self, app=None, db=None):
        self.Blueprint = Blueprint

        # arg mismatch
        if app is not None and \
           db is None and \
           not isinstance(app, Flask):
            self.init_db(app)
            app = None

        # proper spec
        if db is not None:
            self.init_db(db)

        # app is specified properly
        if app is not None:
            self.init_app(app)
        return

    def init_app(self, app, db=None):
        """
        Initialize application via lazy factory pattern.

        Args:
            app (Flask): Flask application.
            db (SQAlchemy): Flask SQLAlchemy extension.
        """
        if db is not None:
            self.init_db(db)

        self.app = app
        self.app.route = types.MethodType(route, self.app)
        self.app.config.setdefault('OCCAM_AUTODOC_ENABLED', True)
        self.app.config.setdefault('OCCAM_AUTODOC_PREFIX', '/docs')
        self.app.config.setdefault('OCCAM_LOG_USER_FORMAT', 'user')
        self.app.config.setdefault('OCCAM_LOG_DEFAULT_LEVEL', 'info')
        self.app.url_map.converters['id'] = ModelConverter
        self.app.register_error_handler(ValidationError, ValidationError.handler)

        # add auto-documentation if specified
        if self.app.config['OCCAM_AUTODOC_PREFIX']:
            @app.route(self.app.config['OCCAM_AUTODOC_PREFIX'] + '/<path:endpoint>')
            def autodoc(endpoint):
                endpoint = '/' + endpoint
                if endpoint in DOCS:
                    doc = []
                    for method in METHODS:
                        if method in DOCS[endpoint]:
                            doc.append(DOCS[endpoint][method])
                    return "<pre>\n" + '\n---\n'.join(doc) + "\n</pre>", 200
                return '', 204

        return

    def init_db(self, db):
        """
        Initialize database model extensions.

        Args:
            db (SQAlchemy): Flask SQLAlchemy extension.
        """
        self.db = db
        self.db.Model.__bases__ += (ModelMixin,)
        return
